@@ -9,6 +9,7 @@ import csrbox.csrgrp_check as csrgrp_check
 import datetime
 import textwrap
 import math
+from riscv_config.warl import warl_interpreter
 
 version = csrbox.__version__
 logger = logging.getLogger(__name__)
@@ -49,6 +50,10 @@ def csr_gen (isa_file, grp_file, custom_file, debug_file, custom_attr_file, work
     global debug_enabled
     global compressed_enabled
     global supervisor_enabled
+    global hypervisor_enabled
+    global physical_addr_sz
+    global pmp_grain
+    global satp_mode_condition
 
     if 'C' in isa_yaml['hart0']['ISA']:
         compressed_enabled = True
@@ -59,6 +64,11 @@ def csr_gen (isa_file, grp_file, custom_file, debug_file, custom_attr_file, work
         supervisor_enabled = True
     else:
         supervisor_enabled = False
+    
+    if 'H' in isa_yaml['hart0']['ISA']:
+        hypervisor_enabled = True
+    else:
+        hypervisor_enabled = False
 
     if custom_file :
         custom_yaml= utils.load_yaml(custom_file)
@@ -71,6 +81,20 @@ def csr_gen (isa_file, grp_file, custom_file, debug_file, custom_attr_file, work
     else:
         debug_enabled = False
 
+    physical_addr_sz = isa_yaml['hart0']['physical_addr_sz']
+    pmp_grain = isa_yaml['hart0']['pmp_granularity']
+    valid_satp_modes=[]
+    xlen = max(isa_yaml['hart0']['supported_xlen'])
+    if supervisor_enabled:
+        satp_modewarl =\
+                    (warl_interpreter(isa_yaml['hart0']['satp']['rv'+str(xlen)]['mode']['type']['warl']))
+        for x in [0, 1, 8, 9]:
+            if satp_modewarl.islegal(x,[]):
+                if xlen == 64:
+                    valid_satp_modes.append('word[63:60] == ' +str(x))
+                else:
+                    valid_satp_modes.append('word[31] == ' +str(x))
+    satp_mode_condition = '||'.join(valid_satp_modes)
     
     csrgrp_check.csrgrp_check(isa_yaml, grp_yaml, logging)
     attr_yaml = csrgrp_check.calculate_size(isa_yaml, custom_attr_file)
@@ -113,6 +137,9 @@ package {0} ;
         c_str = ''
     bsvfile.write(csrreq_struct_temp.format(xlen, c_str))
     bsvfile.write(func_temp.format(xlen))
+    if hypervisor_enabled:
+        bsvfile.write(func_address)
+    bsvfile.write(func_external.format(1))
     bsvfile.write(end_package_temp)
     bsvfile.close()
 
@@ -199,7 +226,6 @@ def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, probeinst, logging=False
         bsvstr +=(end_package_temp)
         bsvfile.write(bsvstr)
         station_isa.clear()  
-    physical_addr_sz = isa_yaml['hart0']['physical_addr_sz']
     bsvstr = csr_top(csr_isa, grp_yaml, attr_yaml, aux_info, rvxlen, 
              'csrbox', logging)
     filename = 'csrbox.bsv'
@@ -366,7 +392,9 @@ def create_concat_vector(mylist, xlen):
             cv += prefix + str(name) + ' ,'
         concat_length += 1
         last_lsb = msb - width
-
+    if last_lsb != -1 :
+       cv += readonly_temp.format(last_lsb + 1)
+       concat_length += 1
     return concat_length, cv[:-1]
 
 
@@ -415,7 +443,7 @@ def derive_wlrl_func(wlrl_list):
     return full_wlrl, act_width
     
 
-def derive_warl_func(field_name, width, warl, csr_dep):
+def derive_warl_func(field_name, width, warl, csr_dep, post_op=''):
     '''
         Function to convert the warl strings in the riscv-config YAML to
         corresponding bluespec code.
@@ -460,7 +488,7 @@ def derive_warl_func(field_name, width, warl, csr_dep):
                 op = '!=' if 'not in' in op else '=='
                 condition_str += ' {0}[{1}] {2} {3}'.format(csr_write, indices, op, val)
                 condition_str += ' && '
-            condition_str = condition_str[:-4] + ' )\n                  '
+            condition_str = condition_str[:-4] + ' )begin\n                  '
         
 
 
@@ -491,17 +519,28 @@ def derive_warl_func(field_name, width, warl, csr_dep):
                         curr_bsv_str += '( x[' + indices + '] == ' + v + ') ||'
                 curr_bsv_str = curr_bsv_str[:-2] + ') && '
             curr_bsv_str = curr_bsv_str[:-3] + ')'
-            curr_bsv_str = curr_bsv_str.replace('0x', "'h") + ')\n'
-            curr_bsv_str += '                        r._write(x);'
+            curr_bsv_str = curr_bsv_str.replace('0x', "'h") + ')begin\n'
+            curr_bsv_str += '                       let _x = x;'
+            if post_op != '':
+                curr_bsv_str += post_op
+            curr_bsv_str = curr_bsv_str + 'r._write(_x);end\n'
         elif bitmask_vals != []:
             indices = bitmask_vals[0][0]
             mask_fixed = bitmask_vals[0][1].split(',')
             mask = mask_fixed[0][2:]
             fixedval = mask_fixed[1][3:]
-            curr_bsv_str = "r._write( (x & 'h" + mask + ") | (~'h" + mask + " & " + fixedval + ") );"
+            curr_bsv_str = "let _x = ( (x & 'h" + mask + ") | (~'h" + mask + " & 'h" + fixedval + ") );"
+            if post_op != '':
+                curr_bsv_str = curr_bsv_str + post_op + ';'
+            curr_bsv_str = curr_bsv_str + 'r._write(_x);'
         if full_warl:
-            curr_bsv_str = 'r._write(x);'
-        bsv_str += condition_str + curr_bsv_str + '\n'
+            curr_bsv_str = 'let _x = x;'
+            if post_op != '':
+                curr_bsv_str = curr_bsv_str + post_op + ';'
+            curr_bsv_str = curr_bsv_str + 'r._write(_x);'
+        if condition_str != '': 
+            curr_bsv_str += ' end\n'
+        bsv_str += condition_str + curr_bsv_str
         bsv_str += '                else  '
     bsv_str = bsv_str[:-22]
     return bsv_str
@@ -540,6 +579,8 @@ def station_interface(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
     global debug_enabled
     global compressed_enabled
     global supervisor_enabled
+    global hypervisor_enabled
+    global physical_addr_sz 
     xlen = int(rvxlen[2:])
     bsvstr += interface_temp.format(pkg_name)
 
@@ -567,6 +608,8 @@ def station_interface(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
             if  logging:
                 logger.debug('-- Creating value methods for ' + csrname)
             size = attr_yaml[csrname]['size']
+            if 'pmpaddr' in csrname:
+                size = physical_addr_sz -2
             bsvstr +=(methodv_dec_temp.format(size, csrname))
 
     # define csr specific action methods.
@@ -678,10 +721,16 @@ def station_module(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
 
     if logging:
         logger.info('Creating module for : ' + pkg_name)
+    xlen = int(rvxlen[2:])
     attr_str = ''
     for csrname in isa_yaml:
         if 'mod_attributes' in attr_yaml[csrname]:
             for modattr in attr_yaml[csrname]['mod_attributes']:
+               if '::' in modattr :
+                        condition, _modattr = re.findall(r'^(.*?)::(.*?)$',modattr)[0]
+                        if eval(condition) :
+                           attr_str += _modattr +'\n' 
+               else:
                 attr_str += '  ' + modattr + '\n'
     bsvstr += (module_temp.format(pkg_name, attr_str, ''))
     if logging:
@@ -900,6 +949,10 @@ def station_req(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
     global debug_enabled
     global compressed_enabled
     global supervisor_enabled
+    global hypervisor_enabled
+    global satp_mode_condition
+    if supervisor_enabled:
+        satp_mode_condn = satp_mode_condition
 
     if logging:
         logger.info('Generating Core request routine for : ' + pkg_name)
@@ -929,12 +982,12 @@ def station_req(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
             write_cond = ''
 
         if is_shadow: 
-            regname = shadow_csr
+            regname = shadow_csr.replace('.','_')
         logger.debug('Case statement for ' +csrname + ' Address: ' +\
                 str(address))
 
         if 'csr_op' in attr_yaml[csrname] and (\
-                shadow_csr is None or shadow_csr in isa_yaml):
+                shadow_csr is None or shadow_csr.split('.')[0] in isa_yaml):
                 for csrop in attr_yaml[csrname]['csr_op']:
                     if '::' in csrop:
                         cond = csrop.split('::')[0]
@@ -948,8 +1001,7 @@ def station_req(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                     else:
                         cust_csrop = csrop
                         break
-                _t = cust_csrop.replace('size',str(size))
-                _t = _t.replace('xlen',str(xlen))
+                _t = cust_csrop.format(**locals())
                 _t = textwrap.indent(_t, '            ') + '\n'
                 bsvstr += _t
         elif ro:
@@ -999,6 +1051,8 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
         :type work_dir: str
     '''
     global debug_enabled
+    global physical_addr_sz
+    global pmp_grain
         
     if logging:
         logger.info('Generating storage elements for : ' + pkg_name)
@@ -1009,6 +1063,15 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
     xlen = int(rvxlen[2:])
     # creating additional required registers
     req_regs = []
+    for csrname in isa_yaml:
+        if 'pmpaddr' in csrname:
+            if pmp_grain == 0:
+                bsvstr += pmp_masks_grain0.format(physical_addr_sz-2)
+            elif pmp_grain == 1:
+                bsvstr += pmp_masks_grain1.format(physical_addr_sz-2)
+            else:
+                bsvstr += pmp_masks_grain2.format(physical_addr_sz-2)
+            break
     for csrname in isa_yaml:
         if csrname in attr_yaml and 'additional_regs' in attr_yaml[csrname.lower()]:
             for add_reg in attr_yaml[csrname.lower()]['additional_regs']:
@@ -1079,7 +1142,14 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
                 bsv_str = derive_warl_func( csrname, str(size),
                         content[rvxlen]['type']['warl']['legal'], 
                         content[rvxlen]['type']['warl']['dependency_fields'])
-                bsvstr += temp_warlReg.format(description, str(size), 
+                if 'pmpaddr' in csrname:
+                    pmpaddrnum = int(csrname.split('pmpaddr')[1],10)
+                    pmpcfgnum = int(pmpaddrnum / int(xlen/8))
+                    bsvstr += temp_warlReg_pmpaddr.format(description,
+                            physical_addr_sz-2, csrname, reset_val, bsv_str, 
+                            pmpcfgnum, pmpaddrnum)
+                else:
+                    bsvstr += temp_warlReg.format(description, str(size), 
                             csrname, reset_val, bsv_str)
             else:
                 bsvstr +=  temp_register.format(description, str(xlen), csrname,
@@ -1120,11 +1190,12 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
                         bsvstr += temp_wire.format(width,csrname+'_'+subfield)
                     elif subcontent['shadow'] is not None:
                         shadow_csr = subcontent['shadow'].split('.')[0]
-                        if shadow_csr not in isa_yaml and shadow_csr not in incomming_reads and\
-                                attr_yaml[shadow_csr]['size']is not None:
-                            bsvstr += temp_wire.format(size, shadow_csr)
-                            incomming_reads.append(shadow_csr)
-
+                        if (shadow_csr not in isa_yaml and #shadow_csr not in incomming_reads #and\
+                                attr_yaml[shadow_csr]['size']is not None)  :
+                            bsvstr += temp_readOnlyRegister.format(subdescription,str(width), shadow_csr+'_'+subfield, 0)
+                            #incomming_reads.append(shadow_csr)
+                        elif shadow_csr not in isa_yaml and attr_yaml[shadow_csr]['size']is None:
+                            bsvstr += temp_readOnlyRegister.format(subdescription,str(width), shadow_csr+'_'+subfield, 0)
                     elif 'ro_constant' in type_reg:
                         bsvstr += temp_readOnlyRegister.format(subdescription,
                                 str(width), csrname+'_'+subfield, subreset_val)
@@ -1164,6 +1235,11 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkg_name,
             length, cv = create_concat_vector(concat, int(xlen))
             bsvstr += concat_temp.format(description, xlen, csrname,\
                       length, cv)
+        if 'reg_dec' in attr_yaml[csrname]['action_method'] :
+            bsvstr += attr_yaml[csrname]['action_method']['reg_dec']
+            if logging:
+                logger.debug('-- Creating external pending interrupt logic for : ' + csrname)
+        
     if logging:
         logger.info('Storage elements created for : ' + pkg_name)
     return bsvstr
@@ -1207,6 +1283,7 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
     '''
     global debug_enabled
     global compressed_enabled
+    global physical_addr_sz
 
     grp_list = sorted(list(grp_yaml))
     xlen = int(rvxlen[2:])
@@ -1220,15 +1297,26 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
 
     misa_grp = find_group(grp_yaml, 'misa')
     mstatus_grp = find_group(grp_yaml, 'mstatus')
+    mstatush_grp = find_group(grp_yaml, 'mstatush')
     mepc_grp = find_group(grp_yaml, 'mepc')
     mtval_grp = find_group(grp_yaml, 'mtval')
+    mtval2_grp = find_group(grp_yaml, 'mtval2')
     mcause_grp = find_group(grp_yaml, 'mcause')
+    mtinst_grp = find_group(grp_yaml, 'mtinst')
     mtvec_grp = find_group(grp_yaml, 'mtvec')
     sstatus_grp = find_group(grp_yaml, 'sstatus')
     sepc_grp = find_group(grp_yaml, 'sepc')
     stval_grp = find_group(grp_yaml, 'stval')
     scause_grp = find_group(grp_yaml, 'scause')
     stvec_grp = find_group(grp_yaml, 'stvec')
+    vsepc_grp = find_group(grp_yaml, 'vsepc')
+    vstval_grp = find_group(grp_yaml, 'vstval')
+    vsstatus_grp = find_group(grp_yaml, 'vsstatus')
+    vscause_grp = find_group(grp_yaml, 'vscause')
+    vstvec_grp = find_group(grp_yaml, 'vstvec')
+    htval_grp = find_group(grp_yaml, 'htval')
+    htinst_grp = find_group(grp_yaml, 'htinst')
+    hstatus_grp = find_group(grp_yaml, 'hstatus')
     ustatus_grp = find_group(grp_yaml, 'ustatus')
     uepc_grp = find_group(grp_yaml, 'uepc')
     utval_grp = find_group(grp_yaml, 'utval')
@@ -1238,6 +1326,8 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
     mideleg_grp = find_group(grp_yaml, 'mideleg')
     sedeleg_grp = find_group(grp_yaml, 'sedeleg')
     sideleg_grp = find_group(grp_yaml, 'sideleg')
+    hedeleg_grp = find_group(grp_yaml, 'hedeleg')
+    hideleg_grp = find_group(grp_yaml, 'hideleg')
     dpc_grp = find_group(grp_yaml,'dpc')
     dcsr_grp_name = find_group(grp_yaml,'dcsr')
     mip_grp_name = find_group(grp_yaml, 'mip')
@@ -1256,9 +1346,10 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
     supervisor = False
     if None not in (sstatus_grp, sepc_grp, stval_grp, scause_grp, stval_grp):
         supervisor = True
-
-
-
+    hypervisor=False
+    if None not in (hstatus_grp, htval_grp, htinst_grp, vsepc_grp, vstval_grp, vscause_grp, vstvec_grp):
+        hypervisor = True
+        
     # imports of each station
     if logging:
         logger.debug('-- Creating imports')
@@ -1277,6 +1368,8 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
             logger.debug('-- Creating Interface for ' + csrname)
         if attr_yaml[csrname]['value_method']['required']:
             size = attr_yaml[csrname]['size']
+            if 'pmpaddr' in csrname:
+                size = physical_addr_sz -2
             bsvstr +=(methodv_dec_temp.format(size, csrname))
     bsvstr += '''  
   endinterface:Sbread'''
@@ -1345,10 +1438,11 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
     mod_attr_str = ''
     if 'minstret' in isa_yaml:
         if isa_yaml['minstret']['rv'+str(xlen)]['accessible']:
-            mod_attr_str = '  (*conflict_free="ma_incr_minstret,sbread.mv_csr_minstret"*)'
-    if 'minstreth' in isa_yaml:
-        if isa_yaml['minstreth']['rv'+str(xlen)]['accessible']:
-            mod_attr_str += '\n  (*conflict_free="ma_incr_minstret,sbread.mv_csr_minstreth"*)'
+            mod_attr_str = '(*conflict_free="ma_incr_minstret,sbread.mv_csr_minstret"*)'
+
+    for i in range(len(grp_list)-1):
+        mod_attr_str += '\n(*mutually_exclusive="mav_upd_on_trap,x{0}_mkConnectionAVtoAf"*)'.format(str(i))
+        mod_attr_str += '\n(*mutually_exclusive="mav_upd_on_ret,x{0}_mkConnectionAVtoAf"*)'.format(str(i))
     extra_reset = ''
     bsvstr += top_module_temp.format(pkgname,mod_attr_str, extra_reset) + '\n'
 
@@ -1360,11 +1454,14 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
    
     bsvstr += '''
     let lv_misa_s = {0}.mv_csr_misa[18];
+    let lv_misa_h = {0}.mv_csr_misa[7];
     let lv_misa_u = {0}.mv_csr_misa[20];
     let lv_misa_n = {0}.mv_csr_misa[13];
     let lv_misa_c = {0}.mv_csr_misa[2];
     /*doc:reg: holds the current privilege level*/
     Reg#(Privilege_mode) rg_prv <- mkReg(Machine);
+    /*doc:reg: holds the current virtual mode */
+    Reg#(Bit#(1)) rg_virtual <- mkReg(0);
 '''.format(misa_grp)
     
     if debug_enabled:
@@ -1404,7 +1501,7 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
     if logging:
         logger.debug('-- Creating connections between core-fwd interfaces')
     for i in range(len(grp_list)-1):
-        bsvstr += temp_req_connection.format(grp_list[i], grp_list[i+1])
+        bsvstr += temp_req_connection.format(grp_list[i], grp_list[i+1], str(i))
     
     # -- connect rg_prv to all stations
     for grp in grp_yaml:
@@ -1412,6 +1509,7 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
         if debug_enabled:
             bsvstr += temp_dbg_connection.format(grp)
 
+  
     # create action methods of dependent csrs
     if logging:
         logger.debug('-- defining action methods')
@@ -1465,32 +1563,52 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
     if logging:
         logger.debug('-- Defining trap and return methods')
     bsvstr += method_top_def_temp
-    bsvstr += method_top_ret_temp.format(xlen)
-    if debug_enabled:
-        bsvstr += method_top_ret_debug_temp.format(dpc_grp, dcsr_grp_name,
-                mstatus_grp);
-    if usertraps :
-        bsvstr += method_top_ret_user_temp.format(mstatus_grp, uepc_grp)
-    if supervisor:
-        bsvstr += method_top_ret_supervisor_temp.format(mstatus_grp, sepc_grp)
-    bsvstr += method_top_ret_machine_temp.format(mstatus_grp, mepc_grp)
 
-    bsvstr += method_top_trap_temp.format(xlen)
-    if usertraps and supervisor and medeleg_grp is not None \
-            and mideleg_grp is not None:
-        bsvstr += method_top_trap_sedeleg.format(sedeleg_grp, sideleg_grp)
-    if (usertraps or supervisor) and medeleg_grp is not None:
-        bsvstr += method_top_trap_medeleg.format(medeleg_grp, mideleg_grp)
     if debug_enabled:
-        bsvstr += method_top_trap_debug_temp.format(dpc_grp, dcsr_grp_name)
-    if supervisor:
-        bsvstr += method_top_trap_supervisor_temp.format(stval_grp, sepc_grp, scause_grp,
-            mstatus_grp, stvec_grp)
-    if usertraps:
-        bsvstr += method_top_trap_user_temp.format(utval_grp, uepc_grp, ucause_grp,
-            mstatus_grp, utvec_grp)
-    bsvstr += method_top_trap_machine_temp.format(mtval_grp, mepc_grp, mcause_grp,
-            mstatus_grp, mtvec_grp)
+        dret_str = dret_temp.format(dpc_grp, dcsr_grp_name, mstatus_grp)
+    else:
+        dret_str = ''
+
+    if not supervisor_enabled:
+        bsvstr += xret_mu.format(xlen, dret_str, mstatus_grp, mepc_grp)
+    elif not hypervisor_enabled:
+        bsvstr += xret_msu.format(xlen, dret_str, mstatus_grp, mepc_grp)
+    else:
+      if xlen == 64:
+        bsvstr += xret_mhsu.format(xlen, dret_str, sepc_grp, vsstatus_grp,
+                mstatus_grp, hstatus_grp, mepc_grp, vsepc_grp)
+      else:
+        bsvstr += xret_mhsu32.format(xlen, dret_str, sepc_grp, vsstatus_grp,
+                mstatus_grp, hstatus_grp, mepc_grp, vsepc_grp, mstatush_grp)
+
+
+    if debug_enabled:
+        debug_trap_str = trap_debug.format(dpc_grp, dcsr_grp_name)
+    else:
+        debug_trap_str = ''
+
+    if not supervisor:
+        bsvstr += trap_mu.format(xlen, mtval_grp, mepc_grp, mcause_grp,
+                mstatus_grp, mtvec_grp, debug_trap_str)
+    elif supervisor and not hypervisor:
+        bsvstr += trap_msu.format(xlen, medeleg_grp, mideleg_grp, stval_grp,
+                sepc_grp, scause_grp, mstatus_grp, stvec_grp, mtval_grp,
+                mepc_grp, mcause_grp, mtvec_grp, debug_trap_str)
+    elif supervisor and hypervisor:
+        if xlen == 64:
+            bsvstr += trap_mhsu.format(xlen, medeleg_grp, mideleg_grp, hedeleg_grp,
+                hideleg_grp, vstval_grp, vsepc_grp, vscause_grp, vstvec_grp,
+                vsstatus_grp, scause_grp, sepc_grp, stval_grp, stvec_grp,
+                htval_grp, htinst_grp, hstatus_grp, mstatus_grp, mtval_grp,
+                mepc_grp, mcause_grp, mtval2_grp, mtinst_grp, mtvec_grp,
+                debug_trap_str)
+        else:
+            bsvstr += trap_mhsu32.format(xlen, medeleg_grp, mideleg_grp, hedeleg_grp,
+                hideleg_grp, vstval_grp, vsepc_grp, vscause_grp, vstvec_grp,
+                vsstatus_grp, scause_grp, sepc_grp, stval_grp, stvec_grp,
+                htval_grp, htinst_grp, hstatus_grp, mstatus_grp, mtval_grp,
+                mepc_grp, mcause_grp, mtval2_grp, mtinst_grp, mtvec_grp,
+                debug_trap_str, mstatush_grp)
     
     # define additional required interfaces
     req_ifc = []
@@ -1535,7 +1653,17 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, aux_info, rvxlen, pkgname, logging =
                 cfg_count += 1
         bsvstr += pmpaddr_def.format(pmp_entries, physical_addr_sz, pmpaddr_str)
         bsvstr += pmpcfg_def.format(pmp_entries, pmpcfg_str)
-
+    if hypervisor:
+        bsvstr += '''
+        method Action ma_core_req(CSRReq req) ; 
+            req.csr_address = fn_address_virtual(req.csr_address, rg_virtual); 
+            grp1.ma_core_req(req); 
+        endmethod '''
+    else:
+        bsvstr += '''
+        method Action ma_core_req(CSRReq req) ; 
+            grp1.ma_core_req(req); 
+        endmethod '''
     # complete the value method for each csr.
     if logging:
         logger.debug('-- Defining value methods')
